import amqp
import json
import socket
from itertools import chain
from .exception import ConnectionError


class Pub(object):
    """
    Manage connections to the broker for the producer.

    :type host: str
    :param host: the host to connect to, which should be 'host[:port]', if
                 no port is specified then 5672 is used.
    """
    def __init__(self, host='localhost', user='guest', password='guest',
                 virtual_host='/', connect_timeout=3, **kwargs):
        connkwargs = {
                'host': host,
                'userid': user,
                'password': password,
                'virtual_host': virtual_host,
                'connect_timeout': connect_timeout,
                }
        self._pool = ConnectionPool(**connkwargs)

    def __del__(self):
        self._pool.disconnect()

    def publish(self, msg, topic, delivery_mode=1, **properties):
        """
        Publish a msg with a specified topic.

        :type msg: str or dict
        :param msg: the message to publish.

        :type topic: str
        :param topic: the specified topic, should be in 'a.b.c' format,
                      for example, 'instalikes.user.login'.
        """
        exchange_name = topic.split('.')[0]
        routing_key = topic

        content_type = 'text/plain'
        if isinstance(msg, dict):
            msg = json.dumps(msg)
            content_type = 'application/json'
        message = amqp.Message(msg, content_type=content_type, delivery_mode=delivery_mode,  **properties)
        pool = self._pool
        connection = pool.get_connection()
        try:
            return connection.publish(message, exchange_name, routing_key)
        except socket.error:
            connection.shutdown()
            try:
                return connection.publish(message, exchange_name, routing_key)
            except socket.error:
                connection.shutdown()
            except ConnectionError:
                pass
            except Exception:
                connection.disconnect()
        except ConnectionError:
            pass
        except Exception:
            connection.disconnect()
        finally:
            pool.release(connection)


class Connection(object):
    "Manage connections to rabbitmq-server"
    description_format = 'Connection<host=%(host)s,virtual_host=%(virtual_host)s>'

    def __init__(self, **kwargs):
        self.connkwargs = kwargs
        self._connection = None
        self._channel = None
        self._exchanges = set()

    def __del__(self):
        try:
            self.disconnect()
        except Exception:
            pass

    def publish(self, msg, exchange_name, routing_key, mandatory=False, immediate=False):
        if self._connection is None or not self._connection.connected:
            self.connect()
        if exchange_name not in self._exchanges:
            self._channel.exchange_declare(exchange_name, 'topic', durable=True, auto_delete=False)
            self._exchanges.add(exchange_name)
        self._channel.basic_publish(msg, exchange_name, routing_key, mandatory, immediate)

    def connect(self):
        try:
            self._connection = amqp.Connection(**self.connkwargs)
            self._connection.connect()
            self._channel = self._connection.channel()
        except Exception:
            raise ConnectionError('Connect failed')

    def disconnect(self):
        if self._connection is None:
            return
        try:
            self._connection.close()
        except socket.error:
            pass
        self._connection = None
        self._channel = None
        self._exchanges.clear()

    def shutdown(self):
        try:
            if self._connection and self._connection.transport:
                self._connection.transport.close()
        except socket.error:
            pass
        self._connection = None
        self._channel = None
        self._exchanges.clear()


class ConnectionPool(object):
    def __init__(self, connection_class=Connection, max_connections=None, **kwargs):
        self.connection_class = connection_class
        self.max_connections = max_connections or 2 ** 31
        self.connkwargs = kwargs
        self._created_connections = 0
        self._available_connections = []
        self._in_use_connections = set()

    def __str__(self):
        return self.connection_class % self.connkwargs

    def make_connection(self):
        if self._created_connections >= self.max_connections:
            raise ConnectionError('Too many connections')
        self._created_connections += 1
        return self.connection_class(**self.connkwargs)

    def get_connection(self):
        try:
            connection = self._available_connections.pop()
        except IndexError:
            connection = self.make_connection()
        self._in_use_connections.add(connection)
        return connection

    def release(self, connection):
        self._in_use_connections.remove(connection)
        self._available_connections.append(connection)

    def disconnect(self):
        all_conns = chain(self._available_connections,
                          self._in_use_connections)
        for connection in all_conns:
            connection.disconnect()




"""
This module provides an interface to handle messages asynchronously,
using rabbitmq as the broker.
"""
from .pub import Pub
from .sub import Sub
from .hub import PubSubHub


__all__ = ['Pub', 'Sub', 'PubSubHub', 'logger']




import json
import socket
import logging
import amqp
import string
import random
from .exception import ConnectionError
from .connection import ConnectionCluster, wait_func
from .rabbit import RabbitConnection
try:
    import urlparse
except ImportError:
    from urllib import parse as urlparse

logger = logging.getLogger(__file__)


def random_string(length):
    return ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(length))


class PubSubHub(object):
    def __init__(self, url=None, queue_group=None, connection_class=RabbitConnection, **kwargs):
        self.queue_group = queue_group or random_string(8)
        params = self.parse_url(url, **kwargs)
        self.pub_conncluster = ConnectionCluster(connection_class=connection_class, max_connections=100, **params)
        self.sub_conncluster = ConnectionCluster(connection_class=connection_class, max_connections=1, **params)

    def __del__(self):
        self.pub_conncluster.disconnect()
        self.sub_conncluster.disconnect()

    @staticmethod
    def parse_url(server_urls, **kwargs):
        if not isinstance(server_urls, (list, tuple)):
            server_urls = [server_urls]

        endpoints = []
        for url in server_urls:
            args = {}
            parsed = urlparse.urlparse(url)
            virtual_host = parsed.path.strip('/') or '/'
            args.update({"host": parsed.hostname, "port": parsed.port,
                         "virtual_host": virtual_host})
            if parsed.username:
                args.update({"user": parsed.username})
            if parsed.password:
                args.update({"password": parsed.password})
            if parsed.query:
                query_args = urlparse.parse_qs(parsed.query)
                query_args = dict(map(lambda x: (x[0], x[1][0]), query_args.items()))
                args.update(query_args)
            if 'socket_connect_timeout' not in args:
                args.update({"socket_connect_timeout": 3})
            endpoints.append(args)
        kwargs.update({"endpoints": endpoints})
        return kwargs

    def publish(self, msg, topic, **kwargs):
        exchange_name = topic.split('.')[0]
        routing_key = topic
        content_type = 'text/plain'
        if isinstance(msg, (list, dict)):
            msg = json.dumps(msg)
            content_type = 'application/json'
        message = amqp.Message(msg, content_type=content_type, delivery_mode=1,  **kwargs)
        num = len(self.pub_conncluster.all_connection_pools())
        for i in range(num):
            try:
                connection = self.pub_conncluster.get_connection()
                if not connection:
                    continue
            except Exception as e:
                logger.error(e)
                continue
            try:
                return connection.publish(message, exchange_name, routing_key)
            except socket.error as e:
                logger.error(e)
                self.pub_conncluster.connection_error(connection)
            except ConnectionError as e:
                logger.error(e)
                self.pub_conncluster.connection_error(connection)
            except Exception as e:
                logger.error(e)
                connection.disconnect()
            finally:
                self.pub_conncluster.release(connection)
        logger.error('PUB FAIL', {'topic': topic, 'msg': msg})

    def subscribe(self, topic, callback=None):
        def decorator(callback_fn):
            for pool in self.sub_conncluster.all_connection_pools():
                connection = pool.get_connection()
                try:
                    connection.subscribe(topic, self.queue_group, callback=callback_fn)
                except:
                    logger.info('subscribe failed to connection<%s> with topic<%s> and queue_group<%s>'%(connection, topic, self.queue_group))
                finally:
                    pool.release(connection)
            return callback_fn
        return decorator(callback) if callback else decorator

    def run(self):
        self.sub_conncluster.start_readloop()

    def join(self, timeout=3):
        wait_func()
        self.sub_conncluster.join(timeout)



import json
import socket
import sys
import os
import ssl
import random
import errno
from six import string_types
from itertools import chain
from .exception import ConnectionError
try:
    from queue import LifoQueue, Empty, Full
except ImportError:
    from Queue import LifoQueue, Empty, Full


if socket.socket.__module__ == "gevent.socket":
# if socket.socket.__module__ == "gevent._socket3":
    import gevent
    import gevent.threading
    spawn_func = gevent.spawn
    sleep_func = gevent.sleep
    kill_func = gevent.kill
    lock_class = gevent.threading.Lock
    wait_func = gevent.wait
else:
    import time
    import threading

    def threading_spawn(func, *args, **kwargs):
        t = threading.Thread(target=func, args=args, kwargs=kwargs)
        t.daemon = True
        t.start()
        return t

    def threading_wait():
        while 1:
            try:
                time.sleep(3)
            except (KeyboardInterrupt, SystemExit):
                sys.exit()
    spawn_func = threading_spawn
    sleep_func = time.sleep
    lock_class = threading.Lock
    wait_func = threading_wait


class Connection(object):
    """
    Manages TCP communication to and from a server
    """
    description_format = "Connection<host:%(host)s,port:%(port)s,id:%(id)s>"

    def __init__(self, host, port, socket_connect_timeout=None, socket_timeout=None, **kwargs):
        self.pid = os.getpid()
        self.host = host
        self.port = int(port)
        self.socket_timeout = socket_timeout
        self.socket_connect_timeout = socket_connect_timeout or socket_timeout
        self._sock = None
        self._description_args = {
            'host': self.host,
            'port': self.port,
            'id': id(self)
        }
        self._connect_callbacks = []
        self.buffer = ''
        self.ssl_conn = False
        if 'ssl_keyfile' in kwargs or 'ssl_certfile' in kwargs:
            self.ssl_conn = True
            self.keyfile = kwargs['ssl_keyfile']
            self.certfile = kwargs['ssl_certfile']
            ssl_cert_reqs = kwargs.get('ssl_cert_reqs') or ssl.CERT_NONE
            ssl_ca_certs = kwargs.get('ssl_ca_certs')
            if isinstance(ssl_cert_reqs, string_types):
                cert_reqs = {
                    'none': ssl.CERT_NONE,
                    'optional': ssl.CERT_OPTIONAL,
                    'required': ssl.CERT_REQUIRED
                }
                if ssl_cert_reqs not in cert_reqs:
                    raise Exception("Invalid SSL Certificate Requirements Flag: %s" % ssl_cert_reqs)
                ssl_cert_reqs = cert_reqs[ssl_cert_reqs]
            self.cert_reqs = ssl_cert_reqs
            self.ca_certs = ssl_ca_certs

    def __repr__(self):
        return self.description_format % self._description_args

    def __del__(self):
        try:
            self.disconnect()
        except Exception:
            pass

    def register_connect_callback(self, callback):
        self._connect_callbacks.append(callback)

    def clear_connect_callbacks(self):
        self._connect_callbacks = []

    def connect(self):
        """
        Connects to the server if not already connected
        """
        if self._sock:
            return
        try:
            sock = self._connect()
        except socket.error:
            e = sys.exc_info()[1]
            raise ConnectionError(self._error_message(e))

        self._sock = sock
        try:
            self.on_connect()
        except:
            # clean up after any error in on_connect
            self.disconnect()
            raise

        # run any user callbacks. right now the only internal callback
        # is for rabbitmq_hub channel/pattern resubscription
        for callback in self._connect_callbacks:
            callback(self)

    def _connect(self):
        """
        Create a TCP socket connection
        """
        # we want to mimic what socket.create_connection does to support
        # ipv4/ipv6, but we want to set options prior to calling
        # socket.connect()
        resources = socket.getaddrinfo(self.host, self.port, socket.AF_INET, socket.SOCK_STREAM)
        if len(resources) == 0:
            raise Exception("getaddrinfo returns an empty list")

        index = random.randint(1, len(resources))
        start = index % len(resources)

        for i in range(len(resources)):
            family, socktype, proto, canonname, socket_address = resources[(start+i) % len(resources)]
            sock = None

            try:
                sock = socket.socket(family, socktype, proto)
                # TCP_NODELAY
                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

                # set the socket_connect_timeout before we connect
                sock.settimeout(self.socket_connect_timeout)

                # connect
                sock.connect(socket_address)

                # set the socket_timeout now that we're connected
                sock.settimeout(self.socket_timeout)

                if self.ssl_conn:
                    sock = ssl.wrap_socket(sock, cert_reqs=self.cert_reqs, keyfile=self.keyfile,
                                           certfile=self.certfile, ca_certs=self.ca_certs)
                return sock

            except socket.error as _:
                err = _
                if sock is not None:
                    sock.close()
                if i == len(resources)-1:
                    raise

        raise socket.error("socket.getaddrinfo returned an empty list")

    def _error_message(self, exception):
        # args for socket.error can either be (errno, "message")
        # or just "message"
        if len(exception.args) == 1:
            return "Error connecting to %s:%s. %s." % \
                (self.host, self.port, exception.args[0])
        else:
            return "Error %s connecting to %s:%s. %s." % \
                (exception.args[0], self.host, self.port, exception.args[1])

    def on_connect(self):
        pass

    def disconnect(self):
        """
        Disconnects from the server
        :return: 
        """
        if self._sock is None:
            return
        try:
            self._sock.shutdown(socket.SHUT_RDWR)
            self._sock.close()
        except socket.error:
            pass
        self._sock = None
        self.on_disconnect()

    def on_disconnect(self):
        pass

    def connection(self):
        if not self._sock:
            self.connect()
        return self._sock

    def _read(self, n=None):
        sock = self.connection()
        return sock.recv(n)

    def read(self, bytes):
        while len(self.buffer) < bytes:
            data = None
            try:
                data = self._read(1024)
            except socket.error as ex:
                if ex.args[0] == errno.EINTR:
                    continue
                raise ex
            if not data:
                break
            self.buffer += data
        result = self.buffer[:bytes]
        self.buffer = self.buffer[bytes:]
        return str(result)

    def write(self, string):
        sock = self.connection()
        try:
            return sock.sendall(string)
        except:
            self.disconnect()
            raise


class ConnectionPool(object):
    def __init__(self, connection_class=Connection, max_connections=None, **kwargs):
        self.connection_class = connection_class
        self.max_connections = max_connections or 2 ** 31
        self.connection_kwargs = kwargs

        self.reset()

    def __repr__(self):
        return "%s<%s>" % (
            type(self).__name__,
            self.connection_class.description_format % self.connection_kwargs,
        )

    def reset(self):
        self.pid = os.getpid()
        self._created_connections = 0
        self._available_connections = []
        self._in_use_connections = set()
        self._check_lock = lock_class()

    def _checkpid(self):
        if self.pid != os.getpid():
            with self._check_lock:
                if self.pid == os.getpid():
                    # another thread already did the work while we waited
                    # on the lock.
                    return
                self.disconnect()
                self.reset()

    def get_connection(self):
        "Get a connection from the pool"
        self._checkpid()
        try:
            connection = self._available_connections.pop()
        except IndexError:
            connection = self.make_connection()
        self._in_use_connections.add(connection)
        return connection

    def make_connection(self):
        "Create a new connection"
        if self._created_connections >= self.max_connections:
            raise ConnectionError("Too many connections")
        self._created_connections += 1
        return self.connection_class(**self.connection_kwargs)

    def release(self, connection):
        "Releases the connection back to the pool"
        self._checkpid()
        if connection.pid != self.pid:
            return
        self._in_use_connections.remove(connection)
        self._available_connections.append(connection)

    def disconnect(self):
        "Disconnects all connections in the pool"
        all_conns = chain(self._available_connections,
                          self._in_use_connections)
        for connection in all_conns:
            connection.disconnect()


class BlockingConnectionPool(ConnectionPool):
    """
    Thread-safe blocking connection pool.
    It performs the same function as the default
    ``:py:class: ~dock.common.connection.ConnectionPool`` implementation, in that,
    it maintains a pool of reusable connections (safely across threads if required).
    The difference is that, in the event that a client tries to get a
    connection from the pool when all of connections are in use, rather than
    raising a ``:py:class: ~dock.common.exceptions.ConnectionError`` (as the default
    ``:py:class: ~dock.common.connection.ConnectionPool`` implementation does), it
    makes the client wait ("blocks") for a specified number of seconds until
    a connection becomes available.
    Use ``max_connections`` to increase / decrease the pool size::
        >>> pool = BlockingConnectionPool(max_connections=10)
    Use ``timeout`` to tell it either how many seconds to wait for a connection
    to become available, or to block forever:
        # Block forever.
        >>> pool = BlockingConnectionPool(timeout=None)
        # Raise a ``ConnectionError`` after five seconds if a connection is
        # not available.
        >>> pool = BlockingConnectionPool(timeout=5)
    """
    def __init__(self, max_connections=100, timeout=6,
                 connection_class=Connection, queue_class=LifoQueue,
                 **connection_kwargs):

        self.queue_class = queue_class
        self.timeout = timeout
        super(BlockingConnectionPool, self).__init__(
            connection_class=connection_class,
            max_connections=max_connections,
            **connection_kwargs)

    def reset(self):
        self.pid = os.getpid()
        self._check_lock = lock_class()

        # Create and fill up a thread safe queue with ``None`` values.
        self.pool = self.queue_class(self.max_connections)
        while True:
            try:
                self.pool.put_nowait(None)
            except Full:
                break

        # Keep a list of actual connection instances so that we can
        # disconnect them later.
        self._connections = []

    def make_connection(self):
        "Make a fresh connection."
        connection = self.connection_class(**self.connection_kwargs)
        self._connections.append(connection)
        return connection

    def get_connection(self):
        """
        Get a connection, blocking for ``self.timeout`` until a connection
        is available from the pool.
        If the connection returned is ``None`` then creates a new connection.
        Because we use a last-in first-out queue, the existing connections
        (having been returned to the pool after the initial ``None`` values
        were added) will be returned before ``None`` values. This means we only
        create new connections when we need to, i.e.: the actual number of
        connections will only increase in response to demand.
        """
        # Make sure we haven't changed process.
        self._checkpid()

        # Try and get a connection from the pool. If one isn't available within
        # self.timeout then raise a ``ConnectionError``.
        # connection = None
        try:
            connection = self.pool.get(block=True, timeout=self.timeout)
        except Empty:
            raise ConnectionError("No connection available.")

        # If the ``connection`` is actually ``None`` then that's a cue to make
        # a new connection to add to the pool.
        if connection is None:
            connection = self.make_connection()

        return connection

    def release(self, connection):
        "Releases the connection back to the pool."
        # Make sure we haven't changed process.
        self._checkpid()
        if connection.pid != self.pid:
            return

        # Put the connection back into the pool.
        try:
            self.pool.put_nowait(connection)
        except Full:
            # perhaps the pool has been reset() after a fork? regardless,
            # we don't want this connection
            pass

    def disconnect(self):
        "Disconnects all connections in the pool."
        for connection in self._connections:
            connection.disconnect()


class ConnectionCluster(object):
    CHECK_INTERVAL = 10

    def __init__(self, connection_class=Connection, connection_pool_class=ConnectionPool, endpoints=[], **kwargs):
        self.connection_class = connection_class
        self.connection_pool_class = connection_pool_class

        self.endpoints = dict(map(lambda x: ('%s:%s' % (x['host'], x['port']), x), endpoints))

        for k, v in self.endpoints.items():
            v['connection_class'] = self.connection_class
            v.update(kwargs)

        self._available_pools = {}
        self._checking_pools = set()
        for k, v in self.endpoints.items():
            self._available_pools[k] = self.connection_pool_class(**v)
        self._processors = []

    def all_connection_pools(self):
        return list(self._available_pools.values())

    def get_connection(self):
        if not self._available_pools:
            return None
        index = random.randint(1, len(self._available_pools)) % len(self._available_pools)
        pool = self.all_connection_pools()[index]
        conn = pool.get_connection()
        try:
            conn.connect()
        except socket.error:
            self.connection_error(conn)
        except ConnectionError:
            self.connection_error(conn)
        return conn

    def release(self, connection):
        key = '%s:%s'%(connection.host, connection.port)
        pool = self._available_pools.get(key)
        if pool:
            pool.release(connection)

    def disconnect(self):
        for k, pool in self._available_pools.items():
            try:
                pool.disconnect()
            except:
                pass

    def join(self, timeout=3):
        for thread in self._processors:
            try:
                thread.join(timeout)
            except (KeyboardInterrupt, SystemExit):
                print('Shutting down ...')

    def connection_error(self, connection):
        connection.disconnect()
        self.release(connection)
        key = '%s:%s'%(connection.host, connection.port)
        if key not in self._checking_pools:
            pool = self._available_pools.pop(key)
            spawn_func(self.check_endpoint, key, pool)
            self._checking_pools.add(key)

    def check_endpoint(self, key, pool):
        while 1:
            conn = pool.get_connection()
            try:
                conn.connect()
                self._available_pools[key] = pool
                self._checking_pools.remove(key)
                break
            except socket.error:
                sleep_func(self.CHECK_INTERVAL)
            except ConnectionError:
                sleep_func(self.CHECK_INTERVAL)
            except:
                sleep_func(self.CHECK_INTERVAL)
            finally:
                pool.release(conn)

    def start_readloop(self):
        funcs = []
        for pool in self._available_pools.values():
            funcs.append(spawn_func(self._conn_readloop, pool))
        self._processors = funcs
        return funcs

    def _conn_readloop(self, pool):
        while 1:
            try:
                connection = pool.get_connection()
            except:
                sleep_func(self.CHECK_INTERVAL)
                continue
            try:
                connection.readloop()
            except socket.error:
                connection.disconnect()
                self._conn_check(connection)
            except ConnectionError:
                connection.disconnect()
                self._conn_check(connection)
            except:
                connection.disconnect()
                self._conn_check(connection)
            finally:
                pool.release(connection)

    def _conn_check(self, connection):
        while 1:
            try:
                connection.connect()
                break
            except socket.error:
                sleep_func(self.CHECK_INTERVAL)
            except ConnectionError:
                sleep_func(self.CHECK_INTERVAL)
            finally:
                sleep_func(self.CHECK_INTERVAL)



import amqp
import json
import socket
import sys, os
from .exception import ConnectionError


class CallbackWrapper(object):
    def __init__(self, topic, callback):
        self.topic = topic
        self._callback = callback

    def __str__(self):
        return ('CallbackWrapper<topic: %s, callback: %s>'
                % (self.topic, self.callback.__name__))

    def callback(self, msg):
        if self._callback is not None:
            if msg.content_type == 'application/json':
                try:
                    message = json.loads(msg.body)
                except Exception:
                    message = msg.body
            else:
                message = msg.body
            self._callback(self.topic, message)


class RabbitConnection(object):
    """
    Manage connections to rabbitmq-server
    """
    description_format = 'RabbitConnection<host=(%(host)s, %(port)s),virtual_host=%(virtual_host)s>'

    def __init__(self, host="127.0.0.1", port=5672, user='guest', password='guest',
                 virtual_host='/', **kwargs):
        self.pid = os.getpid()
        self.host = host
        self.port = port
        self.user, self.password = user, password
        self.virtual_host = virtual_host
        self.connkwargs = {
            'host': self.host, 'userid': self.user,
            'password': self.password,
            'virtual_host': self.virtual_host
        }
        self.connkwargs.update(kwargs)
        self._connection = None
        self._channel = None
        self._exchanges = set()
        self._description_args = {
            'host': self.host, 'port': self.port,
            'virtual_host': self.virtual_host
        }
        self._handlers = {}

    def __str__(self):
        return self.description_format % self._description_args

    def __del__(self):
        try:
            self.disconnect()
        except Exception:
            pass

    def _register_handlers(self):
        for h in self._handlers.values():
            self.subscribe(h['topic'], h['queue_group'], h['callback'])

    @property
    def connection(self):
        if self._connection is None or not self._connection.connected:
            self.connect()
        return self._connection

    @property
    def channel(self):
        if self._connection is None or not self._connection.connected:
            self.connect()
        return self._channel

    def publish(self, msg, exchange_name, routing_key, mandatory=False, immediate=False):
        if exchange_name not in self._exchanges:
            self.channel.exchange_declare(exchange_name, 'topic', durable=True, auto_delete=False)
            self._exchanges.add(exchange_name)
        self.channel.basic_publish(msg, exchange_name, routing_key, mandatory, immediate)

    def subscribe(self, topic, queue_group, callback=None):
        if '*' in topic or '#' in topic:
            print('* and # wildcard will not support in the next rabbitmq_hub. Use the topic fullname directly.')

        self._handlers[topic] = dict(topic=topic, queue_group=queue_group, callback=callback)

        exchange_name = topic.split('.', 1)[0]
        queue_name = '%s.%s' % (topic, queue_group)

        self.channel.exchange_declare(exchange_name, 'topic', durable=True, auto_delete=False)
        self.channel.queue_declare(queue_name, durable=True)
        self.channel.queue_bind(queue_name, exchange_name, topic)

        wrapper = CallbackWrapper(topic, callback)
        self.channel.basic_consume(queue_name, callback=wrapper.callback, no_ack=True)

    def readloop(self):
        try:
            self.connection.drain_events()
        except IOError:
            raise ConnectionError('readloop connetion error.')

    def connect(self):
        if self._connection:
            return
        try:
            self._connection = amqp.Connection(**self.connkwargs)
            self._connection.connect()
            self._channel = self._connection.channel()
            self._register_handlers()
        except Exception:
            raise ConnectionError('Connect failed')

    def disconnect(self):
        if self._connection is None:
            return
        try:
            self._connection.close()
        except socket.error:
            pass
        except IOError:
            pass
        self._connection = None
        self._channel = None
        self._exchanges.clear()



import amqp
import json
import socket
from functools import partial
import time
from .exception import ConnectionError


class Handler(object):
    def __init__(self, exchange_name, queue_name, binding_key, callback):
        self.exchange_name = exchange_name
        self.queue_name = queue_name
        self.binding_key = binding_key
        self._callback = callback

    def __str__(self):
        return ('handler<exchange: %s, queue: %s, binding_key: %s, callback: %s>'
                % (self.exchange_name, self.queue_name, self.binding_key,
                    self.callback.__name__))

    def callback(self, msg):
        if self._callback is not None:
            if msg.content_type == 'application/json':
                try:
                    message = json.loads(msg.body)
                except Exception:
                    message = msg.body
            else:
                message = msg.body
            self._callback(self.binding_key, message)


class Sub(object):
    """
    Manage connections to the broker for the consumer.

    :type name: str
    :param name: the name of the subscriber. If multiple subscribers with the same name
                 subscribe the same topic, messages will be delivered to these subscribers
                 in a round-robin way.

    :type host: str
    :param host: the host to connect to, which should be 'host[:port]', if
                 no port is specified then 5672 is used.
    """
    def __init__(self, name, host='localhost', user='guest', password='guest',
                 virtual_host='/', connect_timeout=None, reconnect_interval=30, **kwargs):
        self.connkwargs = {
                'host': host,
                'userid': user,
                'password': password,
                'virtual_host': virtual_host,
                'connect_timeout': connect_timeout,
                }
        self.reconnect_interval = reconnect_interval
        self._name = name
        self._connection = None
        self._channel = None
        self._handlers = {}

    def __del__(self):
        try:
            self.disconnect()
        except Exception:
            pass

    def _connect(self):
        if self._connection is not None:
            return
        try:
            self._connection = amqp.Connection(**self.connkwargs)
            self._connection.connect()
            self._channel = self._connection.channel()
        except Exception:
            raise ConnectionError('Connect failed')

        self._register_handlers()

    def _register_handlers(self):
        for h in self._handlers.values():
            self._channel.exchange_declare(h.exchange_name, 'topic', durable=True, auto_delete=False)
            self._channel.queue_declare(h.queue_name, durable=True)
            self._channel.queue_bind(h.queue_name, h.exchange_name, h.binding_key)
            self._channel.basic_consume(h.queue_name, callback=partial(h.callback), no_ack=True)

    @property
    def connection(self):
        if not self._connection:
            try:
                self._connect()
            except:
                pass
        return self._connection

    def disconnect(self):
        if self._connection is None:
            return
        try:
            self._connection.close()
        except socket.error:
            pass
        self._connection = None
        self._channel = None

    def shutdown(self):
        try:
            if self._connection and self._connection.transport:
                self._connection.transport.close()
        except socket.error:
            pass
        self._connection = None
        self._channel = None

    def subscribe(self, topic, callback=None):
        """
        Subscribe messages on a specified topic with a specified message handler.

        :type topic: str
        :param topic: the specified topic to subscribe, should be in 'a.b.c' format.
        :type callback: callable object
        :param callback: the messsage handler: callback(topic, message)

                      * topic: the topic you have subscribed to
                      * message: the message deliverd to the topic, str or dict

        """
        if '*' in topic or '#' in topic:
            print('* and # wildcard will not support in the next rabbitmq_hub. Use the topic fullname directly.')
        exchange_name = topic.split('.')[0]
        queue_name = '%s.%s' % (topic, self._name)
        binding_key = topic

        def decorator(callback):
            handler = Handler(exchange_name, queue_name, binding_key, callback)
            self._handlers[topic] = handler
            return callback
        return decorator(callback) if callback else decorator

    def drain_events(self):
        """
        Get a message from the topic you have subscribed and call the callback, it will not return util the callback
        returns
        """
        connection = self.connection
        if connection:
            try:
                connection.drain_events()
            except IOError:
                self.shutdown()
            except:
                self.disconnect()
                raise
        else:
            time.sleep(self.reconnect_interval)

    def run(self):
        """
        Run a infinite loop to handle messages from the topic you have subscribed continously.
        """
        while True:
            self.drain_events()



# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages

__version__ = '1.2.0'
with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


setup(
    name="rabbitmq_hub",
    version=__version__,
    packages=find_packages(),
    include_package_data=True,
    description="rabbitmq connection manager",
    long_description=README,
    author="Mageia",
    author_email="yzg963@gmail.com",
    zip_safe=False,
    license="Proprietary",
    platforms="any",
    install_requires=['amqp'],
    entry_points={},
)




#!/usr/bin/env python
# coding: utf-8
# Author: `mageia`,
# Email: ``,
# Date: `08/05/2017 15:17`
# Description: ''

import time
from unittest import TestCase
from ..rabbitmq_hub import PubSubHub, Pub, Sub


class RabbitMQTestCase(TestCase):
    @staticmethod
    def user_callback(topic, msg):
        print('user_callback: topic: %s, msg: %s' % (topic, msg))

    def get_hub(self):
        return PubSubHub(url='pubsub://leaniot:leaniot@127.0.0.1:5672/', queue_group='test')

    def test_hub_publish(self):
        p = self.get_hub()
        i = 1
        while True:
            msg = "message: %d" % i
            p.publish(msg, 'rabbitmq.test.1')
            p.publish(msg, 'rabbitmq.test.2')
            p.publish(msg, 'rabbitmq.test.3')
            p.publish(msg, 'rabbitmq.test.4')
            i += 1
            print('published %s messages' % (i))
            time.sleep(0.1)

    def test_hub_subscribe(self):
        h = self.get_hub()
        h.subscribe('rabbitmq.test.1', self.user_callback)
        h.subscribe('rabbitmq.test.2', self.user_callback)

        @h.subscribe('rabbitmq.test.3')
        @h.subscribe('rabbitmq.test.4')
        def media_callback(topic, msg):
            print('media_callback: topic: %s, msg: %s' % (topic, msg))
        h.run()
        h.join()

    def test_pub(self):
        p = Pub()
        i = 1

        while True:
            msg = "message: %d" % i
            p.publish(msg, 'rabbitmq.dashboard.index')
            p.publish(msg, 'rabbitmq.device.registered')
            # p.publish(msg, 'rabbitmq.user.logout')
            #
            i += 1
            # msg = {'a': i}
            # p.publish(msg, 'rabbitmq.media.get')
            #
            # msg = ['aaaa', {'bb': time.time(), 'cc': '121231'}]
            # p.publish(msg, 'rabbitmq.media.upload')
            # print('published %s messages' % (i))
            time.sleep(0.1)

    def test_sub(self):
        s = Sub('test', '127.0.0.1', reconnect_interval=10)
        s.subscribe('rabbitmq.user.login', self.user_callback)
        s.subscribe('rabbitmq.user.logout', self.user_callback)

        @s.subscribe('rabbitmq.user.login')
        @s.subscribe('rabbitmq.media.get')
        @s.subscribe('rabbitmq.media.upload')
        def media_callback(topic, msg):
            print('media_callback: topic: %s, msg: %s' % (topic, msg))
        s.run()

